pub mod enums;
pub mod structs;

use std::{fs::OpenOptions, io::Write, mem, os::raw::c_char, ptr};

use enums::*;
use structs::*;
use vulkan_sys::prelude::*;

#[no_mangle]
#[allow(non_snake_case)]
pub extern "C" fn vkNegotiateLoaderLayerInterfaceVersion(
    pVersionStruct: *mut VkNegotiateLayerInterface,
) -> VkResult {
    write_log("======================================================");
    write_log("================= New Negotiation ====================");
    write_log("======================================================");

    unsafe {
        *pVersionStruct = VkNegotiateLayerInterface {
            sType: enums::VkNegotiateLayerStructType::LAYER_NEGOTIATE_INTERFACE_STRUCT,
            pNext: ptr::null_mut(),
            loaderLayerInterfaceVersion: 2,
            pfnGetInstanceProcAddr: Some(get_instance_proc_addr),
            pfnGetDeviceProcAddr: Some(get_device_proc_addr),
            pfnGetPhysicalDeviceProcAddr: None,
        }
    };

    VK_SUCCESS
}

#[no_mangle]
extern "system" fn get_device_proc_addr(
    device: VkDevice,
    function_name: *const c_char,
) -> PFN_vkVoidFunction {
    write_log("vulkan layer device proc addr");

    let func_string = match VkString::try_from(function_name) {
        Ok(func) => func,
        Err(_) => {
            write_log("Err: failed creating string");
            return Functions::Null.convert();
        }
    };

    write_log(format!("requested function: {}", func_string.as_str()));

    match func_string.as_str() {
        "vkCreateDevice" => Functions::CreateDevice(create_device),
        "vkDestroyDevice" => Functions::DestroyDevice(destroy_device),
        "vkCreateInstance" => Functions::CreateInstance(create_instance),
        "vkDestroyInstance" => Functions::DestroyInstance(destroy_instance),

        _ => Functions::Null,
    }
    .convert()
}

#[no_mangle]
extern "system" fn get_instance_proc_addr(
    instance: VkInstance,
    function_name: *const c_char,
) -> PFN_vkVoidFunction {
    write_log("vulkan layer instance proc addr");

    let func_string = match VkString::try_from(function_name) {
        Ok(func) => func,
        Err(_) => {
            write_log("Err: failed creating string");
            return Functions::Null.convert();
        }
    };

    match func_string.as_str() {
        "vkCreateDevice" => Functions::CreateDevice(create_device),
        "vkDestroyDevice" => Functions::DestroyDevice(destroy_device),
        "vkCreateInstance" => Functions::CreateInstance(create_instance),
        "vkDestroyInstance" => Functions::DestroyInstance(destroy_instance),

        _ => Functions::Null,
    }
    .convert()
}

extern "system" fn create_instance(
    create_info: *const VkInstanceCreateInfo,
    allocator: *const VkAllocationCallbacks,
    instance: *mut VkInstance,
) -> VkResult {
    write_log("vulkan layer create instance");

    let chain_info = match VkLayerInstanceCreateInfo::get_chain_info(
        unsafe { &*create_info },
        VK_LAYER_LINK_INFO,
    ) {
        Some(info) => info,
        None => {
            write_log("chain info not found.");
            return VK_ERROR_LAYER_NOT_PRESENT;
        }
    };

    let proc_addr = chain_info.layer_info().next_instance_proc_addr;
    let create_instance = match unsafe {
        mem::transmute::<PFN_vkVoidFunction, Option<PFN_vkCreateInstance>>(proc_addr(
            VkInstance::NULL_HANDLE,
            VkString::new("vkCreateInstance").as_ptr(),
        ))
    } {
        Some(pfn) => pfn,
        None => return VK_ERROR_INITIALIZATION_FAILED,
    };

    chain_info.advance_layer_info();
    let result = create_instance(create_info, allocator, instance);
    if result != VK_SUCCESS {
        return result;
    }

    write_log("successfully created instance.");

    VK_SUCCESS
}

extern "system" fn destroy_instance(instance: VkInstance, allocator: *const VkAllocationCallbacks) {
    write_log("vulkan layer destroy instance");
}

extern "system" fn create_device(
    physical_device: VkPhysicalDevice,
    create_info: *const VkDeviceCreateInfo<'_>,
    allocator: *const VkAllocationCallbacks,
    device: *mut VkDevice,
) -> VkResult {
    write_log("vulkan layer create device");

    VK_SUCCESS
}

extern "system" fn destroy_device(device: VkDevice, allocator: *const VkAllocationCallbacks) {
    write_log("vulkan layer destroy device");
}

fn write_log(msg: impl ToString) {
    let mut file = OpenOptions::new()
        .append(true)
        .create(true)
        .open("/home/michael/rf2_vk_hud.log")
        .unwrap();

    file.write_all(format!("{}\n", msg.to_string()).as_bytes())
        .unwrap();
}
